from typing import List

from langchain_core.pydantic_v1 import BaseModel, Field


class Output(BaseModel):
    synonyms: List[str] = Field(
        description="Synonyms of keywords provided, make every letter lowercase except for the first letter"
    )



from typing import List

from langchain_core.pydantic_v1 import BaseModel, Field


class Output(BaseModel):
    person: List[str] = Field(description="Names of people")
    organization: List[str] = Field(description="Names of companies or organizations")
    location: List[str] = Field(description="Names of locations or places")
    objects: List[str] = Field(description="Names of tangible thigns")



import logging
import os
import sys
import numpy as np

import geocoder
import googlemaps
from ansistrip import ansi_strip
from dotenv import load_dotenv
from llama_index.core import (
    KnowledgeGraphIndex,
    Settings,
    SimpleDirectoryReader,
    StorageContext,
)
from llama_index.core.agent import ReActAgent
from llama_index.core.llms import ChatMessage
from llama_index.core.multi_modal_llms.generic_utils import load_image_urls
from llama_index.core.query_engine import RetrieverQueryEngine
from llama_index.core.retrievers import KnowledgeGraphRAGRetriever
from llama_index.core.tools import FunctionTool
from llama_index.graph_stores.neo4j import Neo4jGraphStore
from llama_index.llms.openai import OpenAI
from llama_index.llms.perplexity import Perplexity
from llama_index.multi_modal_llms.openai import OpenAIMultiModal
import requests

from memary.agent.data_types import Message
from memary.agent.llm_api.tools import openai_chat_completions_request
from memary.memory import EntityKnowledgeStore, MemoryStream
from memary.synonym_expand.synonym import custom_synonym_expand_fn

MAX_ENTITIES_FROM_KG = 5
ENTITY_EXCEPTIONS = ["Unknown relation"]
# ChatGPT token limits
CONTEXT_LENGTH = 4096
EVICTION_RATE = 0.7
NONEVICTION_LENGTH = 5
TOP_ENTITIES = 20


def generate_string(entities):
    cypher_query = "MATCH p = (n) - [*1 .. 2] - ()\n"
    cypher_query += "WHERE n.id IN " + str(entities) + "\n"
    cypher_query += "RETURN p"

    return cypher_query


class Agent(object):
    """Agent manages the RAG model, the ReAct agent, and the memory stream."""

    def __init__(
        self,
        name,
        memory_stream_json,
        entity_knowledge_store_json,
        system_persona_txt,
        user_persona_txt,
        past_chat_json,
        debug=True,
    ):
        load_dotenv()
        # getting necessary API keys
        os.environ["OPENAI_API_KEY"] = os.getenv("OPENAI_API_KEY")
        googlemaps_api_key = os.getenv("GOOGLEMAPS_API_KEY")
        pplx_api_key = os.getenv("PERPLEXITY_API_KEY")

        # Neo4j credentials
        self.neo4j_username = "neo4j"
        self.neo4j_password = os.getenv("NEO4J_PW")
        self.neo4j_url = os.getenv("NEO4J_URL")
        database = "neo4j"

        # initialize APIs
        # OpenAI API
        self.model = "gpt-3.5-turbo"
        self.openai_api_key = os.environ["OPENAI_API_KEY"]
        self.model_endpoint = "https://api.openai.com/v1"
        self.openai_mm_llm = OpenAIMultiModal(
            model="gpt-4-vision-preview",
            api_key=os.getenv("OPENAI_KEY"),
            max_new_tokens=300,
        )
        self.llm = OpenAI(model="gpt-3.5-turbo-instruct")
        self.query_llm = Perplexity(
            api_key=pplx_api_key, model="mistral-7b-instruct", temperature=0.5
        )
        self.gmaps = googlemaps.Client(key=googlemaps_api_key)
        Settings.llm = self.llm
        Settings.chunk_size = 512

        # initialize Neo4j graph resources
        self.graph_store = Neo4jGraphStore(
            username=self.neo4j_username,
            password=self.neo4j_password,
            url=self.neo4j_url,
            database=database,
        )

        self.vantage_key = os.getenv("ALPHA_VANTAGE_API_KEY")
        #self.news_data_key = os.getenv("NEWS_DATA_API_KEY")

        self.storage_context = StorageContext.from_defaults(graph_store=self.graph_store)
        graph_rag_retriever = KnowledgeGraphRAGRetriever(
            storage_context=self.storage_context,
            verbose=True,
            llm=self.llm,
            retriever_mode="keyword",
            synonym_expand_fn=custom_synonym_expand_fn,
        )

        self.query_engine = RetrieverQueryEngine.from_args(
            graph_rag_retriever,
        )

        search_tool = FunctionTool.from_defaults(fn=self.search)
        locate_tool = FunctionTool.from_defaults(fn=self.locate)
        vision_tool = FunctionTool.from_defaults(fn=self.vision)
        stock_tool = FunctionTool.from_defaults(fn=self.stock_price)
        # news_tool = FunctionTool.from_defaults(fn=self.get_news)

        self.debug = debug
        self.routing_agent = ReActAgent.from_tools(
            [search_tool, locate_tool, vision_tool, stock_tool], llm=self.llm, verbose=True
        )

        self.memory_stream = MemoryStream(memory_stream_json)
        self.entity_knowledge_store = EntityKnowledgeStore(entity_knowledge_store_json)

        self.message = Message(
            system_persona_txt, user_persona_txt, past_chat_json, self.model
        )

    def __str__(self):
        return f"Agent {self.name}"

    def external_query(self, query: str):
        messages_dict = [
            {"role": "system", "content": "Be precise and concise."},
            {"role": "user", "content": query},
        ]
        messages = [ChatMessage(**msg) for msg in messages_dict]
        external_response = self.query_llm.chat(messages)

        return str(external_response)
            

    def search(self, query: str) -> str:
        """Search the knowledge graph or perform search on the web if information is not present in the knowledge graph"""
        response = self.query_engine.query(query)

        if response.metadata is None:
            return self.external_query(query)
        else:
            return response

    def locate(self, query: str) -> str:
        """Finds the current geographical location"""
        location = geocoder.ip("me")
        lattitude, longitude = location.latlng[0], location.latlng[1]

        reverse_geocode_result = self.gmaps.reverse_geocode((lattitude, longitude))
        formatted_address = reverse_geocode_result[0]["formatted_address"]
        return "Your address is" + formatted_address

    def vision(self, query: str, img_url: str) -> str:
        """Uses computer vision to process the image specified by the image url and answers the question based on the CV results"""
        img_docs = load_image_urls([img_url])
        response = self.openai_mm_llm.complete(prompt=query, image_documents=img_docs)
        return response
    
    def stock_price(self, query: str) -> str:
        """Get the stock price of the company given the ticker"""
        request_api = requests.get(r'https://www.alphavantage.co/query?function=GLOBAL_QUOTE&symbol=' + query + r'&apikey=' + self.vantage_key)
        return(request_api.json())

    # def get_news(self, query: str) -> str:
    #     """Given a keyword, search for news articles related to the keyword"""
    #     request_api = requests.get(r'https://newsdata.io/api/1/news?apikey=' + self.news_data_key + r'&q=' + query)
    #     return request_api.json()


    def query(self, query: str) -> str:
        # get the response from react agent
        response = self.routing_agent.chat(query)
        self.routing_agent.reset()
        # write response to file for KG writeback
        with open("data/external_response.txt", "w") as f:
            print(response, file=f)
        # write back to the KG
        self.write_back()
        return response

    def write_back(self):
        documents = SimpleDirectoryReader(
            input_files=["data/external_response.txt"]
        ).load_data()

        KnowledgeGraphIndex.from_documents(
            documents,
            storage_context=self.storage_context,
            max_triplets_per_chunk=8,
        )

    def check_KG(self, query: str) -> bool:
        """Check if the query is in the knowledge graph.

        Args:
            query (str): query to check in the knowledge graph

        Returns:
            bool: True if the query is in the knowledge graph, False otherwise
        """
        response = self.query_engine.query(query)

        if response.metadata is None:
            return False
        return generate_string(
            list(list(response.metadata.values())[0]["kg_rel_map"].keys())
        )

    def _select_top_entities(self):
        entity_knowledge_store = self.message.llm_message['knowledge_entity_store']
        entities = [entity.to_dict() for entity in entity_knowledge_store]
        entity_counts = [entity['count'] for entity in entities]
        top_indexes = np.argsort(entity_counts)[:TOP_ENTITIES]
        return [entities[index] for index in top_indexes]

    def _change_llm_message_chatgpt(self) -> dict:
        """Change the llm_message to chatgpt format.

        Returns:
            dict: llm_message in chatgpt format
        """
        llm_message_chatgpt = self.message.llm_message.copy()
        llm_message_chatgpt["messages"] = []
        top_entities = self._select_top_entities()
        logging.info(f"top_eneities: {top_entities}")
        llm_message_chatgpt["messages"].append(
            {
                "role": "user",
                "content": "Knowledge Entity Store:"
                + str(top_entities),
            }
        )
        llm_message_chatgpt["messages"].extend(
            [context.to_dict() for context in self.message.llm_message["messages"]]
        )
        llm_message_chatgpt.pop("knowledge_entity_store")
        llm_message_chatgpt.pop("memory_stream")
        return llm_message_chatgpt

    def _summarize_contexts(self, total_tokens: int):
        """Summarize the contexts.

        Args:
            total_tokens (int): total tokens in the response
        """
        messages = self.message.llm_message["messages"]

        # First two messages are system and user personas
        if len(messages) > 2 + NONEVICTION_LENGTH:
            messages = messages[2:-NONEVICTION_LENGTH]
            del self.message.llm_message["messages"][2:-NONEVICTION_LENGTH]
        else:
            messages = messages[2:]
            del self.message.llm_message["messages"][2:]

        message_contents = [message.to_dict()["content"] for message in messages]

        llm_message_chatgpt = {
            "model": self.model,
            "messages": [
                {
                    "role": "user",
                    "content": "Summarize these previous conversations into 50 words:"
                    + str(message_contents),
                }
            ],
        }
        response, _ = self._get_gpt_response(llm_message_chatgpt)
        content = "Summarized past conversation:" + response
        self._add_contexts_to_llm_message("assistant", content, index=2)
        logging.info(f"Contexts summarized successfully. \n summary: {response}")
        logging.info(f"Total tokens after eviction: {total_tokens*EVICTION_RATE}")

    def _get_gpt_response(self, llm_message_chatgpt: str) -> str:
        """Get response from the GPT model.

        Args:
            llm_message_chatgpt (str): query to get response for

        Returns:
            str: response from the GPT model
        """
        response = openai_chat_completions_request(
            self.model_endpoint, self.openai_api_key, llm_message_chatgpt
        )
        total_tokens = response["usage"]["total_tokens"]
        response = str(response["choices"][0]["message"]["content"])
        return response, total_tokens

    def get_response(self) -> str:
        """Get response from the RAG model.

        Returns:
            str: response from the RAG model
        """
        llm_message_chatgpt = self._change_llm_message_chatgpt()
        response, total_tokens = self._get_gpt_response(llm_message_chatgpt)
        if total_tokens > CONTEXT_LENGTH * EVICTION_RATE:
            logging.info("Evicting and summarizing contexts")
            self._summarize_contexts(total_tokens)

        self.message.save_contexts_to_json()

        return response

    def get_routing_agent_response(self, query, return_entity=False):
        """Get response from the ReAct."""
        response = ""
        if self.debug:
            # writes ReAct agent steps to separate file and modifies format to be readable in .txt file
            with open("data/routing_response.txt", "w") as f:
                orig_stdout = sys.stdout
                sys.stdout = f
                response = str(self.query(query))
                sys.stdout.flush()
                sys.stdout = orig_stdout
            text = ""
            with open("data/routing_response.txt", "r") as f:
                text = f.read()

            plain = ansi_strip(text)
            with open("data/routing_response.txt", "w") as f:
                f.write(plain)
        else:
            response = str(self.query(query))

        if return_entity:
            # the query above already adds final response to KG so entities will be present in the KG
            return response, self.get_entity(self.query_engine.retrieve(query))
        return response

    def get_entity(self, retrieve) -> list[str]:
        """retrieve is a list of QueryBundle objects.
        A retrieved QueryBundle object has a "node" attribute,
        which has a "metadata" attribute.

        example for "kg_rel_map":
        kg_rel_map = {
            'Harry': [['DREAMED_OF', 'Unknown relation'], ['FELL_HARD_ON', 'Concrete floor']],
            'Potter': [['WORE', 'Round glasses'], ['HAD', 'Dream']]
        }

        Args:
            retrieve (list[NodeWithScore]): list of NodeWithScore objects
        return:
            list[str]: list of string entities
        """

        entities = []
        kg_rel_map = retrieve[0].node.metadata["kg_rel_map"]
        for key, items in kg_rel_map.items():
            # key is the entity of question
            entities.append(key)
            # items is a list of [relationship, entity]
            entities.extend(item[1] for item in items)
            if len(entities) > MAX_ENTITIES_FROM_KG:
                break
        entities = list(set(entities))
        for exceptions in ENTITY_EXCEPTIONS:
            if exceptions in entities:
                entities.remove(exceptions)
        return entities
    

    def update_tools(self, updatedTools):
        print("recieved update tools")
        tools = []
        for tool in updatedTools:
            if tool == "Search":
                tools.append(FunctionTool.from_defaults(fn=self.search))
            elif tool == "Location":
                tools.append(FunctionTool.from_defaults(fn=self.locate))
            elif tool == "Vision":
                tools.append(FunctionTool.from_defaults(fn=self.vision))
            elif tool == "Stocks":
                tools.append(FunctionTool.from_defaults(fn=self.stock_price))
            # elif tool == "News":
            #     tools.append(FunctionTool.from_defaults(fn=self.get_news))
        
        self.routing_agent = ReActAgent.from_tools(tools, llm=self.llm, verbose=True)
 



import os
from typing import List

from dotenv import load_dotenv
from langchain.chains import LLMChain
from langchain.prompts import PromptTemplate
from langchain_core.output_parsers import JsonOutputParser
from langchain_openai import OpenAI

from entity_extraction.output import Output


def custom_entity_extract_fn(query: str) -> List[str]:
    load_dotenv()
    llm = OpenAI(openai_api_key=os.getenv("OPENAI_KEY"), temperature=0)
    parser = JsonOutputParser(pydantic_object=Output)

    template = """
    You are an expert entity extraction system. Extract the following entities from the given text:
    
    Format: {format_instructions}

    Text: {query}
    """

    prompt = PromptTemplate(
        template=template,
        input_variables=["query"],
        partial_variables={
            "format_instructions": parser.get_format_instructions()
        },
    )

    chain = prompt | llm | parser
    result = chain.invoke({"query": query})

    l = []
    for category in result:
        for entity in result[category]:
            l.append(entity)

    return l


# testing
# print(custom_entity_extract_fn("who is harry potter"))



import os
import googlemaps
import json
from langchain.chat_models import ChatOpenAI
from langchain.tools import StructuredTool
from langchain.agents import initialize_agent, AgentType
from pydantic import BaseModel, Field
from datetime import datetime
from typing import Optional, Union, List, Tuple, Any 
from dotenv import load_dotenv

load_dotenv()

google_maps_key = os.getenv("google_maps_key")
gmaps = googlemaps.Client(key=google_maps_key)

def retrieve_location(): 
  "Returns latitude and longitude of current location"
  latitude = 37.423304896346345
  longitude = -122.16023584591919
   
  return latitude, longitude

def reverse_geocode(latitude: float, longitude: float): 
  "Converts latitude and longitude into a readable address"
  result = gmaps.reverse_geocode((latitude, longitude))
  if result:
        address = result[0].get('formatted_address', 'No address found')
  else:
    address = 'No address found'
  
  return address

def find_places(query: None, location = None, radius = None, language = None, min_price = None, max_price = None, open_now = False, place_type = None, region = None): 
  params = {
        'query': query,
        'location': location,
        'radius': radius,
        'language': language,
        'min_price': min_price,
        'max_price': max_price,
        'open_now': open_now,
        'type': place_type,
        'region': region
    } 
  
  params = {k: v for k, v in params.items() if v is not None}
  response = gmaps.places(**params)
  
  results = response['results'][:5] # limit to top 5 results 
  
  simplified_results = []
  for place in results: 
    place_info = {
      'place_id': place.get('place_id'), 
      'name': place.get('name'), 
      'location': (place['geometry']['location']['lat'], place['geometry']['location']['lng']), 
      'formatted_address': place.get('formatted_address'), 
      'rating': place.get('rating')
    }
    simplified_results.append(place_info)
    
  return simplified_results

class FindPlacesArgs(BaseModel): 
  query: Optional[str] = Field(None, description = "Text string to search for, e.g., 'restaurant'")
  location: Optional[Union[tuple, list, str]] = Field(None, description = "Latitude and longitude as a tuple or string")
  radius: Optional[int] = Field(None, description = "Search radius in meters")
  language: Optional[str] = Field(None, description = "The language in which to return results")
  min_price: Optional[int] = Field(None, description = "Minimum price level of places to include, from 0 (most affordable) to 4 (most expensive)")
  max_price: Optional[int] = Field(None, description = "Maximum price level of places to include, from 0 (most affordable) to 4 (most expensive)")
  open_now: Optional[bool] = Field(False, description = "Return only those places that are open for business at the time the query is sent")
  place_type: Optional[str] = Field(None, description = "Type of place to restrict results to")
  region: Optional[str] = Field(None, description = "Region code to bias results towards")
  
find_places_tool = StructuredTool(
  func = find_places, 
  name = "Find Places", 
  description = "Finds places based on location and other optional parameters. For each place, returns place ID, name, coordinates as a tuple, formatted address, and rating.", 
  args_schema = FindPlacesArgs
)

def get_directions(origin, destination, mode="walking", waypoints=None, alternatives=False, avoid=None, language=None, units=None, region=None, departure_time=None, arrival_time=None, optimize_waypoints=False, transit_mode=None, transit_routing_preference=None, traffic_model=None): 
  params = {
        'origin': origin,
        'destination': destination,
        'mode': mode,
        'waypoints': waypoints,
        'alternatives': alternatives,
        'avoid': avoid,
        'language': language,
        'units': units,
        'region': region,
        'departure_time': departure_time,
        'arrival_time': arrival_time,
        'optimize_waypoints': optimize_waypoints,
        'transit_mode': transit_mode,
        'transit_routing_preference': transit_routing_preference,
        'traffic_model': traffic_model
    }
    
  # Remove None values
  params = {k: v for k, v in params.items() if v is not None}
    
  # Make the API call
  response = gmaps.directions(**params)
    
  return response  

class DirectionsArgs(BaseModel):
    origin: Union[str, Tuple[float, float]] = Field(..., description="The starting address or latitude/longitude value.")
    destination: Union[str, Tuple[float, float]] = Field(..., description="The ending address or latitude/longitude value.")
    mode: Optional[str] = Field("walking", description="Mode of transport (driving, walking, bicycling, or transit).")
    waypoints: Optional[Union[List[Union[str, Tuple[float, float]]], str]] = Field(None, description="Array of waypoints to alter the route. To route through a location without stopping, prefix it with 'via:', e.g., ['via:San Francisco', 'via:Mountain View'].")
    alternatives: bool = Field(False, description="If True, more than one route may be returned in the response.")
    avoid: Optional[Union[str, List[str]]] = Field(None, description="Features to avoid (tolls, highways, ferries, etc.).")
    language: Optional[str] = Field(None, description="Language for the results.")
    units: Optional[str] = Field(None, description="Unit system for the results (metric or imperial).")
    region: Optional[str] = Field(None, description="Region code, specified as a ccTLD two-character value.")
    departure_time: Optional[Union[int, datetime]] = Field(None, description="Desired time of departure as a timestamp or datetime.")
    arrival_time: Optional[Union[int, datetime]] = Field(None, description="Desired time of arrival as a timestamp or datetime.")
    optimize_waypoints: bool = Field(False, description="Whether to optimize the order of waypoints.")
    transit_mode: Optional[Union[str, List[str]]] = Field(None, description="Preferred modes of transit (bus, subway, train, etc.).")
    transit_routing_preference: Optional[str] = Field(None, description="Preferences for transit routing (less_walking or fewer_transfers).")
    traffic_model: Optional[str] = Field(None, description="Predictive travel time model (best_guess, optimistic, pessimistic).")

def main(): 
  tools = []
  tools.append(StructuredTool.from_function(retrieve_location))
  tools.append(StructuredTool.from_function(reverse_geocode))
  tools.append(find_places_tool)
  
  # TODO: 
  # Missing: place (implement?)
  # Missing: find_place (implement?)
  # Missing: directions 
  # Missing: distance_matrix
  # Missing: vision 
  # Missing: audio
  
  # llm = ChatOpenAI(model="gpt-4")
  # agent_chain = initialize_agent(tools, llm, agent=AgentType.STRUCTURED_CHAT_ZERO_SHOT_REACT_DESCRIPTION, verbose=True)
  # agent_chain.invoke(
  #   {
  #     "input": "Is there a cafe near me?"
  #   }
  # )
  
  latitude = 37.423304896346345
  longitude = -122.16023584591919
  
  response = get_directions(origin=(latitude, longitude), destination="Coupa Cafe Green Library Stanford")
  # print(response[0]['legs'][0]) # 'legs' because walking was specified?
  # print(response[0]['legs'][0].keys()) # keys = distance, duration, end_address, end_location, start_address, start_location, steps, traffic_speed_entry, via_waypoint
  # print(response[0]['legs'][0]['steps'][0].keys()) # keys = distance, duration, end_location, html_instructions, polyline, start_location, travel_mode
  
  # TODO: 
  # places()
    # Only certain 'type' can be accepted. Prompt agent. 
    # Only certain 'regions' can be accepted. 
  
  
   
if __name__ == "__main__": 
  main()
  # Important GMaps documentation: https://googlemaps.github.io/google-maps-services-python/docs/index.html#googlemaps.Client.places

  
 
  




import unittest
import os
from datetime import datetime, timedelta
from src.memory import MemoryStream
from src.memory.types import MemoryItem

class TestMemoryStream(unittest.TestCase):
    def setUp(self):
        self.file_name = "tests/memory/test_memory.json"
        self.memory_stream = MemoryStream(file_name=self.file_name)

    def tearDown(self):
        # Clean up test file after each test
        try:
            os.remove(self.file_name)
        except FileNotFoundError:
            pass

    def test_add_memory(self):
        data = [MemoryItem("test_entity", datetime.now().replace(microsecond=0))]
        self.memory_stream.add_memory(data)
        self.assertEqual(len(self.memory_stream), 1)
        self.assertEqual(self.memory_stream.get_memory()[0], data[0])

    def test_remove_old_memory(self):
        past_date = datetime.now().replace(microsecond=0) - timedelta(days=10)
        self.memory_stream.add_memory([MemoryItem("old_entity", past_date)])
        self.memory_stream.remove_old_memory(5)
        self.assertEqual(len(self.memory_stream.get_memory()), 0)

    def test_save_and_load_memory(self):
        data = [MemoryItem("test_entity", datetime.now().replace(microsecond=0))]
        self.memory_stream.add_memory(data)
        self.memory_stream.save_memory()
        new_memory_stream = MemoryStream(file_name=self.file_name)
        self.assertEqual(len(new_memory_stream), len(self.memory_stream))
        self.assertEqual(new_memory_stream.get_memory(), self.memory_stream.get_memory())

    def test_get_memory_by_index(self):
        data = [MemoryItem("entity1", datetime.now().replace(microsecond=0)), MemoryItem("entity2", datetime.now().replace(microsecond=0))]
        self.memory_stream.add_memory(data)
        self.assertEqual(self.memory_stream.get_memory_by_index(1), data[1])

    def test_remove_memory_by_index(self):
        data = [MemoryItem("entity1", datetime.now().replace(microsecond=0)), MemoryItem("entity2", datetime.now().replace(microsecond=0))]
        self.memory_stream.add_memory(data)
        self.memory_stream.remove_memory_by_index(0)
        self.assertEqual(len(self.memory_stream), 1)
        self.assertEqual(self.memory_stream.get_memory()[0], data[1])

if __name__ == '__main__':
    unittest.main()



import json
import logging
from dataclasses import dataclass


def save_json(filename, data):
    with open(filename, "w") as f:
        json.dump(data, f, indent=4)


@dataclass
class Context:
    """Context class to store the role and content of the message."""
    role: str  # system or user
    content: str

    def __str__(self):
        return f"{self.role}: {self.content} |"

    def to_dict(self):
        return {'role': self.role, 'content': self.content}


class Message:
    """Message class to store the contexts, memory stream and knowledge entity store."""

    def __init__(self, system_persona_txt, user_persona_txt, past_chat_json, model):
        self.past_chat_json = past_chat_json

        self.contexts = []
        self.system_persona = self.load_persona(system_persona_txt)
        self.user_persona = self.load_persona(user_persona_txt)
        self._init_persona_to_messages()
        self.contexts.extend(self.load_contexts_from_json())

        self.llm_message = {
            "model": model,
            "messages": self.contexts,
            "memory_stream": [],
            "knowledge_entity_store": []
        }

        # self.prompt_tokens = count_tokens(self.contexts)

    def __str__(self):
        llm_message_str = f"System Persona: {self.system_persona}\nUser Persona: {self.user_persona}\n"
        for context in self.contexts:
            llm_message_str += f"{str(context)},"
        for memory in self.llm_message["memory_stream"]:
            llm_message_str += f"{str(memory)},"
        for entity in self.llm_message["knowledge_entity_store"]:
            llm_message_str += f"{str(entity)},"
        return llm_message_str

    def _init_persona_to_messages(self):
        """Initializes the system and user personas to the contexts."""
        self.contexts.append(Context("system", self.system_persona))
        self.contexts.append(Context("user", self.user_persona))

    def load_persona(self, persona_txt) -> str:
        """Loads the persona from the txt file.

        Args:
            persona_txt (str): persona txt file path

        Returns:
            str: persona
        """
        try:
            with open(persona_txt, "r") as file:
                persona = file.read()
            return persona
        except FileNotFoundError:
            logging.info(f"{persona_txt} file does not exist.")

    def load_contexts_from_json(self):
        """Loads the contexts from the past chat json file."""
        try:
            with open(self.past_chat_json, "r") as file:
                data_dicts = json.load(file)

            return [Context(**data_dict) for data_dict in data_dicts]
        except:
            logging.info(
                f"{self.past_chat_json} file does not exist. Starts from empty contexts."
            )
            return []

    def save_contexts_to_json(self):
        """Saves the contexts to the json file.
        We don't save the system and user personas (first two messages)
        """
        save_json(self.past_chat_json, [message.to_dict() for message in self.llm_message['messages'][2:]])



import os
from dotenv import load_dotenv
from langchain_community.graphs import Neo4jGraph
from langchain_community.vectorstores import Neo4jVector
from langchain_openai import OpenAIEmbeddings
from langchain_core.pydantic_v1 import BaseModel, Field
from typing import List
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI
from langchain_community.vectorstores.neo4j_vector import remove_lucene_chars


class Entities(BaseModel):
    """Identifying information about entities."""

    names: List[str] = Field(
        ...,
        description="All the person, organization, or business entities that appear in the text",
    )

def structured_retriever(question: str) -> str:
    """
    Collects the neighborhood of entities mentioned
    in the question
    """
    result = ""
    entities = entity_chain.invoke({"question": question})
    for entity in entities.names:
        response = graph.query(
            """CALL db.index.fulltext.queryNodes('entity', $query, 
            {limit:2})
            YIELD node,score
            CALL {
              MATCH (node)-[r:!MENTIONS]->(neighbor)
              RETURN node.id + ' - ' + type(r) + ' -> ' + neighbor.id AS 
              output
              UNION
              MATCH (node)<-[r:!MENTIONS]-(neighbor)
              RETURN neighbor.id + ' - ' + type(r) + ' -> ' +  node.id AS 
              output
            }
            RETURN output LIMIT 50
            """,
            {"query": generate_full_text_query(entity)},
        )
        result += "\n".join([el['output'] for el in response])
    return result

def generate_full_text_query(input: str) -> str:
    """
    Generate a full-text search query for a given input string.

    This function constructs a query string suitable for a full-text
    search. It processes the input string by splitting it into words and 
    appending a similarity threshold (~2 changed characters) to each
    word, then combines them using the AND operator. Useful for mapping
    entities from user questions to database values, and allows for some 
    misspelings.
    """
    full_text_query = ""
    words = [el for el in remove_lucene_chars(input).split() if el]
    for word in words[:-1]:
        full_text_query += f" {word}~2 AND"
    full_text_query += f" {words[-1]}~2"
    return full_text_query.strip()




import json
import logging
from datetime import datetime

from memary.memory import BaseMemory
from memary.memory.types import MemoryItem

logging.basicConfig(level=logging.INFO)


class MemoryStream(BaseMemory):

    def __len__(self):
        """Returns the number of items in the memory."""
        return len(self.memory)

    def init_memory(self):
        """Initializes memory
        self.memory: list[MemoryItem]
        """
        self.load_memory_from_file()
        if self.entity:
            self.add_memory(self.entity)

    @property
    def return_memory(self):
        return self.memory

    def add_memory(self, entities):
        self.memory.extend([
            MemoryItem(str(entity),
                       datetime.now().replace(microsecond=0))
            for entity in entities
        ])

    def get_memory(self) -> list[MemoryItem]:
        return self.memory
    

    def load_memory_from_file(self):
        try:
            with open(self.file_name, 'r') as file:
                self.memory = [
                    MemoryItem.from_dict(item) for item in json.load(file)
                ]
            logging.info(f"Memory loaded from {self.file_name} successfully.")
        except FileNotFoundError:
            logging.info("File not found. Starting with an empty memory.")



# From MemGPT llm_api_tools.py:

import urllib
import logging
import requests


def smart_urljoin(base_url, relative_url):
    """urljoin is stupid and wants a trailing / at the end of the endpoint address, or it will chop the suffix off"""
    if not base_url.endswith("/"):
        base_url += "/"
    return urllib.parse.urljoin(base_url, relative_url)


def openai_chat_completions_request(url, api_key, data):
    """https://platform.openai.com/docs/guides/text-generation?lang=curl"""

    url = smart_urljoin(url, "chat/completions")
    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {api_key}"
    }

    logging.info(f"Sending request to {url}")
    try:
        # Example code to trigger a rate limit response:
        # mock_response = requests.Response()
        # mock_response.status_code = 429
        # http_error = requests.exceptions.HTTPError("429 Client Error: Too Many Requests")
        # http_error.response = mock_response
        # raise http_error

        # Example code to trigger a context overflow response (for an 8k model)
        # data["messages"][-1]["content"] = " ".join(["repeat after me this is not a fluke"] * 1000)

        response = requests.post(url, headers=headers, json=data)
        logging.info(f"response = {response}")
        response.raise_for_status()  # Raises HTTPError for 4XX/5XX status
        response = response.json()  # convert to dict from string
        logging.info(f"response.json = {response}")
        # response = ChatCompletionResponse(**response)  # convert to 'dot-dict' style which is the openai python client default
        return response
    except requests.exceptions.HTTPError as http_err:
        # Handle HTTP errors (e.g., response 4XX, 5XX)
        logging.error(f"Got HTTPError, exception={http_err}, payload={data}")
        raise http_err
    except requests.exceptions.RequestException as req_err:
        # Handle other requests-related errors (e.g., connection error)
        logging.warning(f"Got RequestException, exception={req_err}")
        raise req_err
    except Exception as e:
        # Handle other potential errors
        logging.warning(f"Got unknown Exception, exception={e}")
        raise e



import os
from typing import Dict

from dotenv import load_dotenv
from langchain.prompts import PromptTemplate
from langchain_core.output_parsers import JsonOutputParser
from langchain_openai import OpenAI

from output import Output


def custom_response_evaluator(query: str, response: str) -> Dict[str, int]:
    load_dotenv()
    llm = OpenAI(openai_api_key=os.getenv("OPENAI_KEY"), temperature=0)
    parser = JsonOutputParser(pydantic_object=Output)

    template = """
    You are an expert response judger. Score the provided respose a score from 0 to 1000 with 0 being the worst and 1000 being the best.
    Give separate scores for conciseness, informativeness and accuracy.
    For conciceness, focus on how succinct the response is and how clearly the information is conveyed using minimal words.
    For informativeness, focus on how comprehensive and relevant the contents of the response is to the given query.
    For accuracy, focus on the correctness and decree of preciseness that the response has.
    Hold the scores you give to the highest quality with higher numbers given only to the best of the best.

    Here is an example: Given the query "tell me about harry potter", and the given response Harry Potter is a series of seven fantasy novels written by British author J.K. Rowling. The novels chronicle the life of a young wizard named Harry Potter and his friends Hermione Granger and Ron Weasley, who are all students at the Hogwarts School of Witchcraft and Wizardry.
    The series follows Harry's life from age 11 to 17. In the first book, Harry learns that his parents were killed by the dark wizard Lord Voldemort when Harry was a baby, and that Harry survived the attack with a lightning-shaped scar on his forehead. Harry then begins attending Hogwarts, where he makes friends, encounters professors like Severus Snape, and faces off against Voldemort and his followers.
    The Harry Potter books have been massively popular, spawning a highly successful film series that ran from 2001 to 2011, as well as a play, video games, theme park attractions, and a vast amount of fan fiction.
    The series is considered a landmark of 20th and 21st century children's literature.
    In summary, Harry Potter is a beloved fantasy series that has had a tremendous cultural impact through its books, films, and expanded universe.
    
    The judge should give a score of 300 for conciseness, 600 to informativeness, 900 for accuracy.
    
    Format: {format_instructions}

    Query: {query}
    Response: {response}
    """

    prompt = PromptTemplate(
        template=template,
        input_variables=["query", "response"],
        partial_variables={"format_instructions": parser.get_format_instructions()},
    )

    chain = prompt | llm | parser
    result = chain.invoke({"query": query, "response": response})
    return result



from memary.memory.base_memory import BaseMemory
from memary.memory.memory_stream import MemoryStream
from memary.memory.entity_knowledge_store import EntityKnowledgeStore

__all__ = ["BaseMemory", "MemoryStream", "EntityKnowledgeStore"]



import os
from typing import List

from dotenv import load_dotenv
from langchain.prompts import PromptTemplate
from langchain_core.output_parsers import JsonOutputParser
from langchain_openai import OpenAI

from synonym_expand.output import Output


def custom_synonym_expand_fn(keywords: str) -> List[str]:
    load_dotenv()
    llm = OpenAI(openai_api_key=os.getenv("OPENAI_KEY"), temperature=0)
    parser = JsonOutputParser(pydantic_object=Output)

    template = """
    You are an expert synonym exapnding system. Find synonyms or words commonly used in place to reference the same word for every word in the list:

    Some examples are:
    - a synonym for Palantir may be Palantir technologies or Palantir technologies inc.
    - a synonym for Austin may be Austin texas
    - a synonym for Taylor swift may be Taylor 
    - a synonym for Winter park may be Winter park resort
    
    Format: {format_instructions}

    Text: {keywords}
    """

    prompt = PromptTemplate(
        template=template,
        input_variables=["keywords"],
        partial_variables={"format_instructions": parser.get_format_instructions()},
    )

    chain = prompt | llm | parser
    result = chain.invoke({"keywords": keywords})

    l = []
    for category in result:
        for synonym in result[category]:
            l.append(synonym.capitalize())

    return l


# testing
# print(custom_synonym_expand_fn("[Nvidia]"))



import ml, utils 
from flask import Flask, request, jsonify, send_file, render_template, make_response
from flask.helpers import send_from_directory
from werkzeug.utils import secure_filename
from dotenv import load_dotenv
import json, os, uuid
import googlemaps

load_dotenv()

app = Flask(__name__)
app.config['UPLOAD_FOLDER'] = './uploads'
google_maps_key = os.getenv("google_maps_key")
gmaps = googlemaps.Client(key=google_maps_key)
           
@app.route('/audio/<filename>', methods=['GET'])
def serve_audio(filename):
    return send_from_directory(os.path.join('./audio_files'), filename, mimetype='audio/wav')

@app.route('/process', methods=['POST'])
def process(): 
  try: 
    question = request.form['question']
    
    if "vision" in question: 
      return handle_vision_question(request)
    elif "location" in question: 
      return handle_location_question(request)
    else: 
      return jsonify({"message": "Question type not recognized."}), 400    
    
  except Exception as e: 
    print(f"Caught exception: {e}")  
    return make_response(jsonify({"error": str(e)}), 400)
  
def handle_vision_question(request): 
  print("Handling vision request ...")
  try: 
    if 'file' not in request.files:
            return jsonify({"error": "No file part in the request"}), 400
          
    file = request.files['file']
    question = request.form['question']
    
    if file.filename == '':
      print("No file selected")
      return jsonify({"error": "No file selected"}), 400
    
    if file:
      filename = secure_filename(file.filename)
      filepath = os.path.join(app.config['UPLOAD_FOLDER'], filename)
      file.save(filepath)
      print(f"Saved file to {filepath}")
      
      base64_image = ml.encode_image(filepath)
      
      gpt_response = ml.call_gpt_vision(base64_image, question)
      print("GPT-4 vision response recieved.")
      
      return handle_audio_response(gpt_response)
  
  except Exception as e:
    print(f"Caught exception in handle_vision_question: {e}")
    return jsonify({"error": str(e)}), 500 

def handle_location_question(request): 
  print("Handling location request ...")
  
  try: 
    latitude = request.form.get('latitude')
    longitude = request.form.get('longitude')
    print(f"Location recieved: ({latitude}, {longitude})")
    
    if not latitude or not longitude: 
      text = "Location could not be determined."
    else: 
      reverse_geocode_result = gmaps.reverse_geocode((latitude, longitude))
      readable_address = reverse_geocode_result[0]['formatted_address']
    
      text = "Your readable address is" + readable_address  
    
    return handle_audio_response(text)
  
  except Exception as e: 
    print(f"Caught exception in handle_location_question: {e}")
    return jsonify({"error": str(e)}), 500 
  
def handle_audio_response(text): 
  audio_directory = './audio_files'  
  if not os.path.exists(audio_directory):
    os.makedirs(audio_directory)
    
  audio_filename = f"audio_{uuid.uuid4()}.mp3"
  audio_path = os.path.join(audio_directory, audio_filename)
  ml.text_to_speech(text, audio_path)
  print(f"Audio file saved at {audio_path}")
    
  audio_url = request.url_root + "audio/" + audio_filename
  return jsonify({"audio_url": audio_url})

@app.route('/')
def home():
    return render_template('index.html')

if __name__ == "__main__": 
  app.run(debug=True)
  # app.run(host='0.0.0.0', port=5000)



import unittest
import os
from datetime import datetime
from src.memory import EntityKnowledgeStore
from src.memory.types import KnowledgeMemoryItem, MemoryItem


class TestEntityKnowledgeStore(unittest.TestCase):

    def setUp(self):
        self.file_name = "tests/memory/test_knowledge_memory.json"
        self.entity_knowledge_store = EntityKnowledgeStore(
            file_name=self.file_name)

    def tearDown(self):
        # Clean up test file after each test
        try:
            os.remove(self.file_name)
        except FileNotFoundError:
            pass

    def test_add_memory(self):
        data = [
            MemoryItem("test_entity",
                       datetime.now().replace(microsecond=0))
        ]
        self.entity_knowledge_store.add_memory(data)
        assert len(self.entity_knowledge_store.knowledge_memory) == 1
        assert isinstance(self.entity_knowledge_store.knowledge_memory[0],
                          KnowledgeMemoryItem)

    def test_convert_memory_to_knowledge_memory(self):
        data = [
            MemoryItem("test_entity",
                       datetime.now().replace(microsecond=0))
        ]
        converted_data = self.entity_knowledge_store._convert_memory_to_knowledge_memory(
            data)
        assert len(converted_data) == 1
        assert isinstance(converted_data[0], KnowledgeMemoryItem)

    def test_update_knowledge_memory(self):
        data = [
            KnowledgeMemoryItem("knowledge_entity", 1,
                                datetime.now().replace(microsecond=0))
        ]
        self.entity_knowledge_store._update_knowledge_memory(data)
        assert len(self.entity_knowledge_store.knowledge_memory) == 1
        assert self.entity_knowledge_store.knowledge_memory[0] == data[0]



import json
import logging

from abc import ABC, abstractmethod
from datetime import datetime, timedelta


class BaseMemory(ABC):

    def __init__(self, file_name: str, entity: str = None):
        """Initializes the memory storage."""
        self.file_name = file_name
        self.entity = entity
        self.memory = []
        self.knowledge_memory = []
        self.init_memory()

    @abstractmethod
    def __len__(self):
        """Returns the number of items in the memory."""
        pass

    @abstractmethod
    def init_memory(self):
        """Initializes memory."""
        pass

    @abstractmethod
    def load_memory_from_file(self):
        """Loads memory from a file."""
        pass

    @abstractmethod
    def add_memory(self, data):
        """Adds new memory data."""
        pass

    @abstractmethod
    def get_memory(self):
        pass

    @property
    def return_memory(self):
        return self.memory

    def remove_old_memory(self, days):
        """Removes memory items older than a specified number of days."""
        cutoff_date = datetime.now() - timedelta(days=days)
        self.memory = [
            item for item in self.return_memory if item.date >= cutoff_date
        ]
        logging.info("Old memory removed successfully.")

    def save_memory(self):
        if self.file_name:
            with open(self.file_name, 'w') as file:
                json.dump([item.to_dict() for item in self.return_memory],
                          file,
                          default=str,
                          indent=4)
                logging.info(f"Memory saved to {self.file_name} successfully.")
        else:
            logging.info("No file name provided. Memory not saved.")

    def get_memory_by_index(self, index):
        if 0 <= index < len(self.memory):
            return self.memory[index]
        else:
            return None

    def remove_memory_by_index(self, index):
        if 0 <= index < len(self.memory):
            del self.memory[index]
            logging.info("Memory item removed successfully.")
            return True
        else:
            logging.info("Invalid index. Memory item not removed.")
            return False

    def clear_memory(self):
        self.memory = []
        if self.file_name:
            with open(self.file_name, 'w') as file:
                json.dump([], file, indent=4) 
                logging.info(f"Memory cleared and saved to {self.file_name} successfully.")
        else:
            logging.info("No file name provided. Memory not cleared or saved.")



from langchain_core.pydantic_v1 import BaseModel, Field


class Output(BaseModel):
    conciseness: int = Field(
        description="Score for how concise the response is from 0 to 1000 witih 0 being the worst and 1000 being the best."
    )
    informativeness: int = Field(
        description="Score for how informative the response is from 0 to 1000 witih 0 being the worst and 1000 being the best."
    )
    accuracy: int = Field(
        description="Score for how accurate the response is based on the question from 0 to 1000 witih 0 being the worst and 1000 being the best."
    )



import tiktoken
from typing import Optional, List

from memary.agent.base_agent import Agent
from memary.agent.data_types import Context


class ChatAgent(Agent):

    def __init__(self, name, memory_stream_json, entity_knowledge_store_json,
                 system_persona_txt, user_persona_txt, past_chat_json):
        super().__init__(name, memory_stream_json, entity_knowledge_store_json,
                         system_persona_txt, user_persona_txt, past_chat_json)
        

    def add_chat(self,
                 role: str,
                 content: str,
                 entities: Optional[List[str]] = None):
        """Add a chat to the agent's memory.

        Args:
            role (str): 'system' or 'user'
            content (str): content of the chat
            entities (Optional[List[str]], optional): entities from Memory systems. Defaults to None.
        """
        # Add a chat to the agent's memory.
        self._add_contexts_to_llm_message(role, content)

        if entities:
            self.memory_stream.add_memory(entities)
            self.memory_stream.save_memory()
            self.entity_knowledge_store.add_memory(
                self.memory_stream.get_memory())
            self.entity_knowledge_store.save_memory()

        self._replace_memory_from_llm_message()
        self._replace_eks_to_from_message()

    def get_chat(self):
        return self.contexts

    def clearMemory(self):
        self.memory_stream.clear_memory()
        self.entity_knowledge_store.clear_memory()
        
       # print("removed from mem stream and entity knowdlege store ")
        "clears knowledge neo4j database"

        print("Deleting nodes from Neo4j...")
        try:
            self.graph_store.query("MATCH (n) DETACH DELETE n")
        except Exception as e:
            print(f"Error deleting nodes: {e}")
        print("Nodes deleted from Neo4j.")



    def _add_contexts_to_llm_message(self, role, content, index=None):
        """Add contexts to the llm_message."""
        if index:
            self.message.llm_message["messages"].insert(index, Context(
                role, content))
        else:
            self.message.llm_message["messages"].append(Context(role, content))

    def _replace_memory_from_llm_message(self):
        """Replace the memory_stream from the llm_message."""
        self.message.llm_message[
            "memory_stream"] = self.memory_stream.get_memory()

    def _replace_eks_to_from_message(self):
        """Replace the entity knowledge store from the llm_message.
        eks = entity knowledge store"""

        self.message.llm_message[
            "knowledge_entity_store"] = self.entity_knowledge_store.get_memory(
            )



from dataclasses import dataclass, asdict
from datetime import datetime


@dataclass
class MemoryItem:
    entity: str
    date: datetime

    def __str__(self):
        return f"{self.entity}, {self.date.isoformat()}"

    def to_dict(self):
        return {'entity': self.entity, 'date': str(self.date.isoformat())}

    @classmethod
    def from_dict(cls, data):
        return cls(entity=data['entity'],
                   date=datetime.fromisoformat(data['date']))


@dataclass
class KnowledgeMemoryItem:
    entity: str
    count: int
    date: datetime

    def __str__(self):
        return f"{self.entity}, {self.count}, {self.date.isoformat()}"

    def to_dict(self):
        return {
            'entity': self.entity,
            'count': self.count,
            'date': str(self.date.isoformat())
        }

    @classmethod
    def from_dict(cls, data):
        return cls(entity=data['entity'],
                   count=data['count'],
                   date=datetime.fromisoformat(data['date']))



import os
import random
import sys
import textwrap

import pandas as pd
import streamlit as st
import streamlit.components.v1 as components
from dotenv import load_dotenv
from neo4j import GraphDatabase
from pyvis.network import Network

# src should sit in the same level as /streamlit_app
curr_dir = os.getcwd()

parent_dir = os.path.dirname(curr_dir)
#parent_dir = os.path.dirname(curr_dir) + '/memary' #Use this if error: src not found. Also move the '/streamlit_app/data' folder to the 'memary' folder, outside the 'src' folder.

sys.path.append(parent_dir + '/src')

from memary.agent.chat_agent import ChatAgent

load_dotenv()

system_persona_txt = "data/system_persona.txt"
user_persona_txt = "data/user_persona.txt"
past_chat_json = "data/past_chat.json"
memory_stream_json = "data/memory_stream.json"
entity_knowledge_store_json = "data/entity_knowledge_store.json"
chat_agent = ChatAgent(
    "Personal Agent",
    memory_stream_json,
    entity_knowledge_store_json,
    system_persona_txt,
    user_persona_txt,
    past_chat_json,
)


def create_graph(nodes, edges):
    g = Network(
        notebook=True,
        directed=True,
        cdn_resources="in_line",
        height="500px",
        width="100%",
    )

    for node in nodes:
        g.add_node(node, label=node, title=node)
    for edge in edges:
        # assuming only one relationship type per edge
        g.add_edge(edge[0], edge[1], label=edge[2][0])

    g.repulsion(
        node_distance=200,
        central_gravity=0.12,
        spring_length=150,
        spring_strength=0.05,
        damping=0.09,
    )
    return g


def fill_graph(nodes, edges, cypher_query):
    entities = []
    with GraphDatabase.driver(
        uri=chat_agent.neo4j_url,
        auth=(chat_agent.neo4j_username, chat_agent.neo4j_password),
    ) as driver:
        with driver.session() as session:
            result = session.run(cypher_query)
            for record in result:
                path = record["p"]
                rels = [rel.type for rel in path.relationships]

                n1_id = record["p"].nodes[0]["id"]
                n2_id = record["p"].nodes[1]["id"]
                nodes.add(n1_id)
                nodes.add(n2_id)
                edges.append((n1_id, n2_id, rels))
                entities.extend([n1_id, n2_id])


cypher_query = "MATCH p = (:Entity)-[r]-()  RETURN p, r LIMIT 1000;"
answer = ""
external_response = ""
st.title("memary Demo")
clear_memory = st.button("Clear Memory DB")
query = st.text_input("Ask a question")



tools = st.multiselect( 
    "Select tools to include:",
    # ["Search", "Location", "Vision", "Stocks", "News"], #all options available
    # ["Search", "Location", "Vision", "Stocks", "News"],) #options that are selected by default


    ["Search", "Location", "Vision", "Stocks"], #all options available
    ["Search", "Location", "Vision", "Stocks"],) #options that are selected by default

if 'Vision' in tools:
    img_url = st.text_input("URL of image, leave blank if no image to provide")
    if img_url:
        st.image(img_url, caption="Uploaded Image", use_column_width=True)


generate_clicked = st.button("Generate")
st.write("")


if clear_memory:
    #print("Front end recieved request to clear memory")
    chat_agent.clearMemory()
    st.write("Memory DB cleared")

if generate_clicked:

    if(query == ""):
        st.write("Please enter a question")
        st.stop()

    #get tools
    print("tools enabled: ", tools)
    if(len(tools) == 0):
        st.write("Please select at least one tool")
        st.stop()

    print("start update tools")
    chat_agent.update_tools(tools)

    if img_url:
        query += "Image URL: " + img_url
    react_response = ""
    rag_response = (
        "There was no information in knowledge_graph to answer your question."
    )
    chat_agent.add_chat("user", query)
    cypher_query = chat_agent.check_KG(query)
    if cypher_query:
        rag_response, entities = chat_agent.get_routing_agent_response(
            query, return_entity=True
        )
        chat_agent.add_chat("user", "rag: " + rag_response, entities)
    else:
        # get response
        react_response = chat_agent.get_routing_agent_response(query)
        chat_agent.add_chat("user", "ReAct agent: " + react_response)

    answer = chat_agent.get_response()
    st.subheader("Routing Agent Response")
    routing_response = ""
    with open("data/routing_response.txt", "r") as f:
        routing_response = f.read()
    st.text(str(routing_response))

    if cypher_query:
        nodes = set()
        edges = []  # (node1, node2, [relationships])
        fill_graph(nodes, edges, cypher_query)

        st.subheader("Knoweldge Graph")
        st.code("# Current Cypher Used\n" + cypher_query)
        st.write("")
        st.text("Subgraph:")
        graph = create_graph(nodes, edges)
        graph_html = graph.generate_html(f"graph_{random.randint(0, 1000)}.html")
        components.html(graph_html, height=500, scrolling=True)
    else:
        st.subheader("Knowledge Graph")
        st.text("No information found in the knowledge graph")

    st.subheader("Final Response")
    wrapped_text = textwrap.fill(answer, width=80)
    st.text(wrapped_text)

    if len(chat_agent.memory_stream) > 0:
        # Memory Stream
        memory_items = chat_agent.memory_stream.get_memory()
        memory_items_dicts = [item.to_dict() for item in memory_items]
        df = pd.DataFrame(memory_items_dicts)
        st.write("Memory Stream")
        st.dataframe(df)

        # Entity Knowledge Store
        knowledge_memory_items = chat_agent.entity_knowledge_store.get_memory()
        knowledge_memory_items_dicts = [
            item.to_dict() for item in knowledge_memory_items
        ]
        df_knowledge = pd.DataFrame(knowledge_memory_items_dicts)
        st.write("Entity Knowledge Store")
        st.dataframe(df_knowledge)

        # top entities
        top_entities = chat_agent._select_top_entities()
        df_top = pd.DataFrame(top_entities)
        st.write("Top 20 Entities")
        st.dataframe(df_top)




import os
from collections import defaultdict
from typing import List

from dotenv import load_dotenv
from llama_index.core.llms import ChatMessage
from llama_index.llms.perplexity import Perplexity
from pandas import DataFrame

from judge import custom_response_evaluator


def benchmark(
    model_name: str,
    data_points: List,
    conciseness: List,
    informativeness: List,
    accuracy: List,
):
    load_dotenv()
    pplx_api_key = os.getenv("PERPLEXITY_API_KEY")

    query_llm = Perplexity(api_key=pplx_api_key, model=model_name, temperature=0.5)
    base_prompt = "tell me about "

    for subject in data_points:
        query = base_prompt + subject
        messages_dict = [
            {"role": "system", "content": "Be precise and concise."},
            {"role": "user", "content": query},
        ]
        messages = [ChatMessage(**msg) for msg in messages_dict]
        response = query_llm.chat(messages)
        result = custom_response_evaluator(query, response)

        conciseness.append(result["conciseness"])
        informativeness.append(result["informativeness"])
        accuracy.append(result["accuracy"])



from openai import OpenAI
import os
import requests
import base64
from dotenv import load_dotenv

load_dotenv()
OPENAI_API_KEY = os.getenv("openai_api_key")

client = OpenAI()
      
def call_gpt_model(prompt, data, model, temperature=None):
  messages = [
      {"role": "system", "content": prompt},
      {"role": "user", "content": data}
  ]

  api_params = {
      "model": model,
      "messages": messages
  }

  if temperature is not None:
    api_params["temperature"] = temperature

  try:
    response = client.chat.completions.create(**api_params)
    response_content = response.choices[0].message.content.strip()

    return response_content

  except Exception as e:
    raise RuntimeError(f"An error occurred while making an API call: {e}")
  
def call_gpt_vision(base64_image, user): 
  headers = {
    "Content-Type": "application/json",
    "Authorization": f"Bearer {OPENAI_API_KEY}"
  }
  
  payload = {
    "model": "gpt-4-vision-preview",
    "messages": [
      {
        "role": "system", 
        "content": "You are tasked with answering a blind individual's question about their current environment. Aim for brevity without sacrificing the immersive experience."
      },
      {
        "role": "user",
        "content": [
          {
            "type": "text",
            "text": user
          },
          {
            "type": "image_url",
            "image_url": {
              "url": f"data:image/jpeg;base64,{base64_image}"
            }
          }
        ]
      }
    ],
    "max_tokens": 300
  }
  
  response = requests.post("https://api.openai.com/v1/chat/completions", headers=headers, json=payload)
  
  return response.json()['choices'][0]['message']['content']

def encode_image(image_path):
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode('utf-8')  

def text_to_speech(text, filepath):   
  try: 
    response = client.audio.speech.create(
      model="tts-1",
      voice="shimmer",
      input=text
    )
    response.stream_to_file(filepath)
    
  except Exception as e: 
    raise RuntimeError(f"An unexpected error occurred: {str(e)}")
    
def speech_to_text(filepath):
  try: 
    audio_file = open(filepath, "rb")
    transcript = client.audio.transcriptions.create(
      model="whisper-1", 
      file=audio_file, 
      prompt="The transcript is about a blind person asking about their environment",
      response_format="text"
    )
  except Exception as e: 
    return f"An unexpected error occurred: {str(e)}"


import json
import logging

from memary.memory import BaseMemory
from memary.memory.types import KnowledgeMemoryItem, MemoryItem


class EntityKnowledgeStore(BaseMemory):

    def __len__(self):
        """Returns the number of items in the memory."""
        return len(self.knowledge_memory)

    def init_memory(self):
        """Initializes memory.
        self.entity_memory: list[EntityMemoryItem]
        """
        self.load_memory_from_file()
        if self.entity:
            self.add_memory(self.entity)

    @property
    def return_memory(self):
        return self.knowledge_memory


    def load_memory_from_file(self):
        try:
            with open(self.file_name, 'r') as file:
                self.knowledge_memory = [
                    KnowledgeMemoryItem.from_dict(item)
                    for item in json.load(file)
                ]
            logging.info(
                f"Entity Knowledge Memory loaded from {self.file_name} successfully."
            )
        except FileNotFoundError:
            logging.info(
                "File not found. Starting with an empty entity knowledge memory."
            )

    def add_memory(self, memory_stream: list[MemoryItem]):
        """To add new memory to the entity knowledge store
        we should convert the memory to knowledge memory and then update the knowledge memory

        Args:
            memory_stream (list): list of MemoryItem
        """
        knowledge_meory = self._convert_memory_to_knowledge_memory(
            memory_stream)
        self._update_knowledge_memory(knowledge_meory)


    def _update_knowledge_memory(self, knowledge_memory: list):
        """update self.knowledge memory with new knowledge memory items

        Args:
            knowledge_memory (list): list of KnowledgeMemoryItem
        """
        for item in knowledge_memory:
            for i, entity in enumerate(self.knowledge_memory):
                if entity.entity == item.entity:
                    self.knowledge_memory[i].date = item.date
                    self.knowledge_memory[i].count += item.count
                    break
            else:
                self.knowledge_memory.append(item)

    def _convert_memory_to_knowledge_memory(
            self, memory_stream: list) -> list[KnowledgeMemoryItem]:
        """Converts memory to knowledge memory

        Returns:
            knowledge_memory (list): list of KnowledgeMemoryItem
        """
        knowledge_memory = []

        entities = set([item.entity for item in memory_stream])
        for entity in entities:
            memory_dates = [
                item.date for item in memory_stream if item.entity == entity
            ]
            knowledge_memory.append(
                KnowledgeMemoryItem(entity, len(memory_dates),
                                    max(memory_dates)))
        return knowledge_memory

    def get_memory(self) -> list[KnowledgeMemoryItem]:
        return self.knowledge_memory

    
    
